/*-
 * Copyright 2015-2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Authors: Nils Goroll <nils.goroll@uplex.de>
 *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <errno.h>

#include "vmod_convert.h"

#include "vrt.h"
#include "vas.h"

#define ILLEGAL -1
#define PAD -2

struct b64_state_s {
	unsigned	magic;
#define B64_STATE_MAGIC	0xc0b64b64
	unsigned char	bytes[2];
	ssize_t	l;
};

static struct b64_alphabet {
	char *b64;
	char i64[256];
	int padding;
} b64_alphabet[__MAX_ENCODING];

static void
alpha_init(struct b64_alphabet *alpha)
{
	int i;
	const char *p;

	for (i = 0; i < 256; i++)
		alpha->i64[i] = ILLEGAL;
	for (p = alpha->b64, i = 0; *p; p++, i++)
		alpha->i64[(int)*p] = (char)i;
	if (alpha->padding)
		alpha->i64[alpha->padding] = PAD;
}

static inline int
decode(char *restrict *restrict dest, char *restrict const buf,
       const size_t maxlen, unsigned u, const int n)
{
	char *d;

	if (n <= 1)
		return -1;
	d = *dest;
	for (int i = 0; i < n - 1; i++) {
		if (d == buf + maxlen)
			return -1;
		*d++ = (u >> 16) & 0xff;
		u <<= 8;
	}
	*dest += d - *dest;
	return 1;
}

void
base64_init(void)
{
	b64_alphabet[BASE64].b64 =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		"ghijklmnopqrstuvwxyz0123456789+/";
	b64_alphabet[BASE64].padding = '=';
	b64_alphabet[BASE64URL].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URL].padding = '=';
	b64_alphabet[BASE64URLNOPAD].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URLNOPAD].padding = 0;
	alpha_init(&b64_alphabet[BASE64]);
	alpha_init(&b64_alphabet[BASE64URL]);
	alpha_init(&b64_alphabet[BASE64URLNOPAD]);
}

/*
 * Base64-encode *in (size: inlen) into the blob supplied as rblob. If
 * there is insufficient space, it will bail out and return
 * NULL. Otherwise, it will null-terminate and return the blob.
 * The enc argument specifies the alphabet and whether padding is used.
 * Inspired heavily by gnulib/Simon Josefsson (as referenced in RFC4648)
 *
 * XXX: tmp[] and idx are used to ensure the reader (and author) retains
 * XXX: a limited amount of sanity. They are strictly speaking not
 * XXX: necessary, if you don't mind going crazy.
 */
ssize_t
base64_encode(const enum encoding enc, char *restrict const buf,
	      const size_t maxlen, const char *restrict const inbuf,
	      const size_t inlength)
{
	struct b64_alphabet *alpha = &b64_alphabet[enc];
	char *p = buf;
	const char *in = inbuf;
	size_t outlen = maxlen, inlen = inlength;
	unsigned char tmp[3], idx;

	AN(buf);
	if (in == NULL || inlen == 0)
		return 0;
	if ((enc != BASE64URLNOPAD && outlen < base64_encode_l(inlen))
	    || outlen < base64nopad_encode_l(inlen))
		return -1;

	while (1) {
		assert(inlen);
		assert(outlen > 3);

		tmp[0] = (unsigned char) in[0];
		tmp[1] = (unsigned char) in[1];
		tmp[2] = (unsigned char) in[2];

		*p++ = alpha->b64[(tmp[0] >> 2) & 0x3f];

		idx = (tmp[0] << 4);
		if (inlen > 1)
			idx += (tmp[1] >> 4);
		idx &= 0x3f;
		*p++ = alpha->b64[idx];

		if (inlen > 1) {
			idx = (tmp[1] << 2);
			if (inlen > 2)
				idx += tmp[2] >> 6;
			idx &= 0x3f;

			*p++ = alpha->b64[idx];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		if (inlen > 2) {
			*p++ = alpha->b64[tmp[2] & 0x3f];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		/*
		 * XXX: Only consume 4 bytes, but since we need a fifth for
		 * XXX: NULL later on, we might as well test here.
		 */
		assert(outlen >= 5);

		outlen -= 4;

		if (inlen < 4)
			break;

		inlen -= 3;
		in += 3;
	}

	assert(outlen > 0);
	*p++ = '\0';
	return p - buf;
}

ssize_t
base64_decode(const enum encoding dec, char *restrict const buf,
	      const size_t maxlen, const char *restrict const p, va_list ap)
{
	struct b64_alphabet *alpha = &b64_alphabet[dec];
	char *dest = buf;
	unsigned u = 0, term = 0;
	int n = 0;

	AN(buf);

	for (const char *s = p; s != vrt_magic_string_end;
	     s = va_arg(ap, const char *)) {
		if (s == NULL)
			continue;
		if (*s && term) {
			errno = EINVAL;
			return -1;
		}
		while (*s) {
			while (n < 4) {
				char b = alpha->i64[(unsigned) *s++];
				u <<= 6;
				if (b == ILLEGAL) {
					errno = EINVAL;
					return -1;
				}
				n++;
				if (b == PAD) {
					term++;
					continue;
				}
				u |= (unsigned) b;
				if (!*s)
					break;
			}
			if (n == 4) {
				if (decode(&dest, buf, maxlen, u, n-term) < 0) {
					errno = ENOMEM;
					return -1;
				}
				n = 0;
			}
		}
	}
	if (n) {
		if (!alpha->padding)
			u <<= 6 * (4 - n);
		if (decode(&dest, buf, maxlen, u, n-term) < 0) {
			errno = ENOMEM;
			return -1;
		}
	}

	return dest - buf;
}
